// This is a script that parses Synchronet log files to generate door stats.

"use strict";

require("sbbsdefs.js", "K_NOCRLF");

var gLogsDir = backslash(system.logs_dir + "logs");

/*
var doorStats = compileAllDoorStats();
printConsoleOrNot("Door stats by door:");
for (var doorName in  doorStats.byDoor)
{
	printConsoleOrNot(doorName + ":");
	printConsoleOrNot(" # times ran: " + doorStats.byDoor[doorName].numTimesRan);
	printConsoleOrNot(" Last run login time: " + doorStats.byDoor[doorName].lastRunLoginDateStr);
	printConsoleOrNot(" Last run by: " + doorStats.byDoor[doorName].lastRunUsername);
	printConsoleOrNot(" Most common user: " + doorStats.byDoor[doorName].mostCommonUser);
	printConsoleOrNot(" User run counts:");
	for (var username in doorStats.byDoor[doorName].usernameRunCounts)
		printConsoleOrNot("  " + username + ": " + doorStats.byDoor[doorName].usernameRunCounts[username]);
}
printConsoleOrNot("");
printConsoleOrNot("Door stats by user:");
for (var username in doorStats.byUsername)
{
	printConsoleOrNot(username + ":");
	printConsoleOrNot(" Total # doors ran: " + doorStats.byUsername[username].totalNumDoorsRan);
	printConsoleOrNot(" Last login time: " + doorStats.byUsername[username].lastLoginDateStr);
	printConsoleOrNot(" Most common door: " + doorStats.byUsername[username].mostCommonDoor);
	printConsoleOrNot(" Door run counts:");
	for (var doorName in doorStats.byUsername[username].doorRunCounts)
		printConsoleOrNot("  " + doorName + ": " + doorStats.byUsername[username].doorRunCounts[doorName]);
}
*/



/////////////////////////////////////////
// Functions

function printConsoleOrNot(pText)
{
	if (typeof(console) === "object" && typeof(console.print) === "function")
		console.print(pText + "\r\n");
	else
		print(pText);
}

// Determines whether 2 strings are equal, ignoring case
function strsEqualIgnoringCase(pStr1, pStr2)
{
	// TODO: This still seems to be case-sensitive
	return pStr1.localeCompare(pStr2, undefined, { sensitivity: 'base' }) === 0;
}

// Returns whether a string is in an array of strings, ignoring case
function strInArrayIgnoringCase(pArrayOfStrs, pStr)
{
	var strFound = false;
	for (var i = 0; i < pArrayOfStrs.length && !strFound; ++i)
		strFound = strsEqualIgnoringCase(pStr, pArrayOfStrs[i]);
	return strFound;
}

// Parses a Synchronet node log file
//
// Parameters:
//  pLogFilename: The name of a Synchronet log file
//  pDoorNamesToSkip: Optional - An array of names of doors to not include
//  pUsernamesToSkip: Optional - An array of usernames to not include
//
// Return value: An array of objects containing the following properties:
//               loginDateTimeStr: The user's login date & time (string)
//               loginDate: A Date object representing the user's login date & time
//               nodeNum: The user's node number (number)
//               userNum: The user number (number)
//               username: The name of the user (string)
//               doorName: The name of the door (string)
function parseLogFile(pLogFilename, pDoorNamesToSkip, pUsernamesToSkip)
{
	var doorEntries = [];

	var inLogFile = new File(pLogFilename);
	if (inLogFile.open("r"))
	{
		var doorLogEntry = null;
		var lastTimeStr = "";
		var lastDateStr = "";
		var lastDate = null;
		var lastNodeNum = -1;
		var lastUserNum = -1;
		var lastUsername = "";
		var regexMatches = null;
		while (!inLogFile.eof)
		{
			// Read the next line from the log file
			var fileLine = inLogFile.readln(2048);

			// fileLine should be a string, but I've seen some cases
			// where for some reason it isn't.  If it's not a string,
			// then continue onto the next line.
			if (typeof(fileLine) != "string")
				continue;
			// If the line is blank, then skip it.
			if (fileLine.length == 0)
				continue;

			// Start of new log entry
			if (fileLine.match(/^@  [0-9][0-9]:[0-9][0-9].*[0-9][0-9][0-9][0-9]/) != null)
			{
				//          1         2         3         4
				//01234567890123456789012345678901234567890123456789
				//@  07:20p  Sun May 29 2022            Node   1
				lastTimeStr = fileLine.substr(3, 6);
				lastNodeNum = +(fileLine.substr(45));
				lastDateStr = fileLine.substr(11, 15);
				// Figure out the time and create a Date object for the last date.
				// The last character in the time string will be "a" for AM or "p" for PM.
				// If it's p, then add 12 to the hours.
				var lastTimeChar = lastTimeStr.charAt(lastTimeStr.length-1).toLowerCase();
				var hoursNum = +(lastTimeStr.substr(0, 2));
				if (lastTimeChar == "p" && hoursNum < 12)
					hoursNum += 12;
				var timeStr = format("%02d:", hoursNum) + lastTimeStr.substr(3, 2);
				lastDate = new Date(lastDateStr + " " + timeStr);
			}
			else if ((regexMatches = fileLine.match(/^\+\+ \(([0-9]+)\)  (.*).*Logon.*/)) != null)
			{
				// There should be 3 matches (the first is the whole line)
				if (regexMatches.length >= 3)
				{
					lastUserNum = +(regexMatches[1].trim());
					lastUsername = regexMatches[2].trim();
				}
			}
			else if (fileLine.indexOf("X- running external program: ") == 0)
			{
				var thisDoorName = fileLine.substr(29).trim();
				var addIt = true;
				if (pDoorNamesToSkip != null)
					addIt = !strInArrayIgnoringCase(pDoorNamesToSkip, thisDoorName);
				if (addIt && pUsernamesToSkip != null)
					addIt = !strInArrayIgnoringCase(pUsernamesToSkip, lastUsername);
				if (addIt)
				{
					doorEntries.push({
						loginDateTimeStr: lastDateStr + " " + lastTimeStr,
						loginDate: lastDate,
						nodeNum: lastNodeNum,
						userNum: lastUserNum,
						username: lastUsername,
						doorName: thisDoorName
					});
				}
			}
		}

		inLogFile.close();
	}

	return doorEntries;
}

// Compiles door stats from all available Synchronet node logs
//
// Parameters:
//  pDoorNamesToSkip: Optional - An array of names of doors to not include
//  pUsernamesToSkip: Optional - An array of usernames to not include
function compileAllDoorStats(pDoorNamesToSkip, pUsernamesToSkip)
{
	var stats = {
		startDate: null,
		endDate: null,
		byDoor: {},
		byUsername: {},
		doorsByPopularity: [],
		usersByMostDoorCount: []
	};

	var logFilenames = directory(gLogsDir + "*.log");
	for (var logFileI = 0; logFileI < logFilenames.length; ++logFileI)
	{
		var doorEntries = parseLogFile(logFilenames[logFileI], pDoorNamesToSkip, pUsernamesToSkip);
		for (var entryI = 0; entryI < doorEntries.length; ++entryI)
		{
			if (stats.startDate == null)
				stats.startDate = doorEntries[entryI].loginDate;
			else
			{
				if (doorEntries[entryI].loginDate < stats.startDate)
					stats.startDate = doorEntries[entryI].loginDate;
			}
			if (stats.endDate == null)
				stats.endDate = doorEntries[entryI].loginDate;
			else
			{
				if (doorEntries[entryI].loginDate > stats.endDate)
					stats.endDate = doorEntries[entryI].loginDate;
			}

			// Update the by-door stats
			if (!stats.byDoor.hasOwnProperty(doorEntries[entryI].doorName))
			{
				stats.byDoor[doorEntries[entryI].doorName] = {
					numTimesRan: 0,
					lastRunLoginDateStr: "",
					lastRunUsername: "",
					usernameRunCounts: {}
				};
			}
			stats.byDoor[doorEntries[entryI].doorName].numTimesRan += 1;
			stats.byDoor[doorEntries[entryI].doorName].lastRunLoginDateStr = doorEntries[entryI].loginDateTimeStr;
			stats.byDoor[doorEntries[entryI].doorName].lastRunLoginDate = doorEntries[entryI].loginDate;
			stats.byDoor[doorEntries[entryI].doorName].lastRunUsername = doorEntries[entryI].username;
			if (!stats.byDoor[doorEntries[entryI].doorName].usernameRunCounts.hasOwnProperty(doorEntries[entryI].username))
				stats.byDoor[doorEntries[entryI].doorName].usernameRunCounts[doorEntries[entryI].username] = 0;
			stats.byDoor[doorEntries[entryI].doorName].usernameRunCounts[doorEntries[entryI].username] += 1;

			// Update the by-username stats
			if (!stats.byUsername.hasOwnProperty(doorEntries[entryI].username))
			{
				stats.byUsername[doorEntries[entryI].username] = {
					totalNumDoorsRan: 0,
					lastLoginDateStr: "",
					doorRunCounts: {}
				};
			}
			stats.byUsername[doorEntries[entryI].username].totalNumDoorsRan += 1;
			stats.byUsername[doorEntries[entryI].username].lastLoginDateStr = doorEntries[entryI].loginDateTimeStr;
			stats.byUsername[doorEntries[entryI].username].lastLoginDate = doorEntries[entryI].loginDate;
			if (!stats.byUsername[doorEntries[entryI].username].doorRunCounts.hasOwnProperty(doorEntries[entryI].doorName))
				stats.byUsername[doorEntries[entryI].username].doorRunCounts[doorEntries[entryI].doorName] = 0;
			stats.byUsername[doorEntries[entryI].username].doorRunCounts[doorEntries[entryI].doorName] += 1;
		}
	}

	// Figure out the most common users for each door
	for (var doorName in stats.byDoor)
	{
		stats.byDoor[doorName].mostCommonUser = "";
		var doorRuns = 0;
		for (var username in stats.byDoor[doorName].usernameRunCounts)
		{
			if (stats.byDoor[doorName].usernameRunCounts[username] > doorRuns)
			{
				doorRuns = stats.byDoor[doorName].usernameRunCounts[username];
				stats.byDoor[doorName].mostCommonUser = username;
			}
		}
	}

	// Figure out the most-often run door per user
	for (var username in stats.byUsername)
	{
		stats.byUsername[username].mostCommonDoor = "";
		var doorRuns = 0;
		for (var doorName in stats.byUsername[username].doorRunCounts)
		{
			if (stats.byUsername[username].doorRunCounts[doorName] > doorRuns)
			{
				doorRuns = stats.byUsername[username].doorRunCounts[doorName];
				stats.byUsername[username].mostCommonDoor = doorName;
			}
		}
	}

	// Populate the stats.doorsByPopularity array
	for (var doorName in stats.byDoor)
	{
		stats.doorsByPopularity.push({
			doorName: doorName,
			numTimesRan: stats.byDoor[doorName].numTimesRan
		});
	}
	stats.doorsByPopularity.sort(function(doorA, doorB) {
		if (doorA.numTimesRan > doorB.numTimesRan)
			return -1;
		else if (doorA.numTimesRan < doorB.numTimesRan)
			return 1;
		else
			return 0;
	});

	// Populate the stats.usersByMostDoorCount array
	for (var username in stats.byUsername)
	{
		var totalNumDoorRuns = 0;
		for (var doorName in stats.byUsername[username].doorRunCounts)
			totalNumDoorRuns += stats.byUsername[username].doorRunCounts[doorName];
		stats.usersByMostDoorCount.push({
			username: username,
			totalNumDoorsRan: stats.byUsername[username].totalNumDoorsRan,
			mostCommonDoor: stats.byUsername[username].mostCommonDoor,
			totalNumDoorRuns: totalNumDoorRuns,
			doorRunCounts: stats.byUsername[username].doorRunCounts
		});
	}
	stats.usersByMostDoorCount.sort(function(userA, userB) {
		if (userA.totalNumDoorRuns > userB.totalNumDoorRuns)
			return -1;
		else if (userA.totalNumDoorRuns < userB.totalNumDoorRuns)
			return 1;
		else
			return 0;
	});

	return stats;
}
